;; The sum of the primes below 10 is 2 + 3 + 5 + 7 = 17.
;;
;; Find the sum of all the primes below one million.
;;

(defvar *some-primes*
  '(2 3 5 7 11 13 17 19))

(defvar *not-primes*
  '(0 1))

(defvar *max-some-primes*
  (apply #'max *some-primes*))

;; should we even assert?
(defun add-cached-prime (prime)
  (when (not (member prime *some-primes*))
    (push prime *some-primes*)
    (setf *max-some-primes* (apply #'max *some-primes*))))

(defun reset-cache ()
  (setf *max-some-primes* 2)
  (setf *some-primes* '(2 3 5)))

(defun range (start end)
  (loop for nn from start to end
	collect nn))

(defun prime? (num)
  (when (member num *not-primes*)
    (return-from prime? nil))
  (when (member num *some-primes*)
    (return-from prime? t))
  (loop for prm in *some-primes*
	do
	(when (= 0 (mod num prm))
	  (return-from prime? nil)))
  (loop for fact = (ceiling (sqrt num))
	then (if (oddp fact)
		 (- fact 2)
		 (- fact 1))
	while (>= fact *max-some-primes*)
	do
	(when (= 0 (mod num fact))
	  (return-from prime? nil)))
  t)

;; lets do some caching...
;; (reset-cache)
;; (mapcar #'add-cached-prime (remove-if-not #'prime? (range 1 10000)))
;; (setf *some-primes* (sort *some-primes* #'<))
;; *max-some-primes*

(defmacro aprog1 (first &body body)
  `(let ((it ,first))
    (progn
      ,@body)
    it))

; (macroexpand-1 '(aprog1 1 2))

(defun make-prime-spigot ()
  (let ((cached-primes *some-primes*)
	(last-num nil))
    (sort cached-primes #'<)
    (lambda ()
      (cond ((not (null cached-primes))
	     (aprog1
		 (pop cached-primes)
	       (setf last-num it)))
	    (t
	     (loop for next = (+ last-num 2)
		   then (+ next 2)
		   do
		   (when (prime? next)
		     (setf last-num next)
		     (return next))))))))

;; (let ((sp (make-prime-spigot)))
;;   (loop for nn from 1 to 100
;; 	do
;; 	(format t "next:~a~&" (funcall sp))))

(time
 (let ((total 0)
       (sp (make-prime-spigot)))
   (loop for nn = (funcall sp)
	 then (funcall sp)
	 while (< nn 1000000)
	 do
	 (setf total (+ total nn)))
   total))

;; 37550402023